@@ -2,9 +2,21 @@
 # coding: utf-8
 # pylint: disable=C0114,C0413,R0902,C0209
 # determine available python executable
+# determine ps -o options
 _=''''
-[ "$1" = -W ] && shift && exec watch -x -c -- "$0" -C y "$@"
-export PGT_PGREP=$(type -p pgrep)
+#[ "$1" = -W ] && shift && exec watch -x -c -- "$0" -C y "$@"
+export LANG=en_US.UTF-8 PYTHONUTF8=1 PYTHONIOENCODING=utf8
+PGT_PGREP=$(type -p pgrep)
+ps -p $$ -o ucomm >/dev/null 2>&1 && PGT_COMM=ucomm
+[ ! "$PGT_COMM" ] && ps -p $$ -o comm >/dev/null 2>&1 && PGT_COMM=comm
+[ "$PGT_COMM" ] && {
+    ps -p $$ -o stime >/dev/null 2>&1 && PGT_STIME=stime
+    [ ! "$PGT_STIME" ] && ps -p $$ -o start >/dev/null 2>&1 && PGT_STIME=start
+    [ ! "$PGT_STIME" ] && PGT_STIME=time
+}
+# busybox no -p option
+[ ! "$PGT_COMM" ] && ! ps -p $$ >/dev/null 2>&1 && PGT_COMM=comm && PGT_STIME=time
+export PGT_COMM PGT_STIME PGT_PGREP
 python=$(type -p python || type -p python3 || type -p python2)
 [ "$python" ] && exec $python "$0" "$@"
 echo "ERROR: cannot find python interpreter" >&2
@@ -17,7 +29,7 @@ Like pstree but with searching for specific processes with pgrep first and displ
 hierarchy of matching processes (parents and children)
 should work on any Unix supporting commands :
 # pgrep
-# ps -e -o pid,ppid,comm,args
+# ps ax -o pid,ppid,comm,args
 (RedHat/CentOS/Fedora/Ubuntu/Suse/Solaris...)
 Compatible python 2 / 3
 
@@ -45,22 +57,30 @@ __license__ = "MIT"
 
 import sys
 import os
-import getopt
 import platform
+import getopt
 import re
+try:
+    import time
+except ImportError:
+    pass
 
-# pylint: disable=E0602
-# pylint: disable=E1101
-if sys.version_info < (3, 0):
-    reload(sys)
-    sys.setdefaultencoding('utf8')
+# impossible detection using ps for AIX/MacOS
+# stime is not start time of process
+system = platform.system()
+PS_OPTION = 'ax'
+if system in ['AIX', 'Darwin']:
+    os.environ['PGT_STIME'] = 'start'
+elif system == 'SunOS': # ps ax -o not supported
+    PS_OPTION = '-e'
+    os.environ['PGT_COMM'] = 'fname' # comm header width not respected
 
 def runcmd(cmd):
     """run command"""
-    pipe = os.popen('"' + '" "'.join(cmd) + '"', 'r')
+    pipe = os.popen(cmd, 'r')
     std_out = pipe.read()
-    pipe.close()
-    return std_out.rstrip('\n')
+    res = pipe.close()
+    return res, std_out.rstrip('\n')
 
 def ask(prompt):
     """input text"""
@@ -118,7 +138,7 @@ class Proctree:
 
     # pylint: disable=R0913
     def __init__(self, use_uid=False, use_ascii=False, use_color=False,
-                 pid_zero=True, optfields=None):
+                 pid_zero=True, opt_fields=None, threads=False):
         """constructor"""
         self.pids = []
         self.ps_info = {}        # ps command info stored
@@ -127,50 +147,84 @@ class Proctree:
         self.pids_tree = {}
         self.top_parents = []
         self.treedisp = Treedisplay(use_ascii, use_color)
-        self.ps_fields = self.get_fields(optfields, use_uid)
+        self.ps_fields = self.get_fields(opt_fields, use_uid, threads)
         self.get_psinfo(pid_zero)
 
-    def get_fields(self, optfields=None, use_uid=False):
+    def get_fields(self, opt_fields=None, use_uid=False, threads=False):
         """ Get ps fields from OS / optionnal fields """
-        osname = platform.system()
-        if optfields:
-            opt_fields = optfields.split(',')
-        else:
-            if osname in ['AIX', 'Darwin']:
-                opt_fields = ['start']
-            else:
-                opt_fields = ['stime']
         if use_uid:
             user = 'uid'
         else:
             user = 'user'
-        if osname == 'SunOS':
-            comm = 'comm'
+        if threads:
+            pid = 'spid'
         else:
-            comm = 'ucomm'
+            pid = 'pid'
+        if not opt_fields or not os.environ.get('PGT_COMM'):
+            opt_fields = [os.environ.get('PGT_STIME') or 'stime']
 
-        return ['pid', 'ppid', user, comm] + opt_fields
+        return [pid, 'ppid', user, os.environ.get('PGT_COMM') or 'ucomm'] + opt_fields
+
+    def run_ps(self, widths):
+        """
+            run ps command detected setting columns widths
+            guess columns for ps command not supporting -o (mingw/msys2)
+        """
+        if os.environ.get('PGT_COMM'):
+            ps_cmd = 'ps ' + PS_OPTION + ' ' + ' '.join(
+                    ['-o '+ o +'='+ widths[i]*'-' for i,o in enumerate(self.ps_fields)]
+                ) + ' -o args'
+            err, ps_out = runcmd(ps_cmd)
+            if err:
+                print('Error: executing ps ' + PS_OPTION + ' -o ' + ",".join(self.ps_fields))
+                sys.exit(1)
+            return ps_out.splitlines()
+        _, out = runcmd('ps aux') # try to use header to guess columns
+        out = out.splitlines()
+        if not 'PPID' in out[0]:
+            _, out = runcmd('ps -ef')
+            out = out.splitlines()
+        ps_out = []
+        fields = {}
+        for i,field in enumerate(out[0].strip().lower().split()):
+            field = re.sub("command|cmd", "args", field)
+            field = re.sub("uid", "user", field)
+            fields[field] = i
+        if not 'ppid' in fields:
+            print("Error: command 'ps aux' does not provides PPID")
+            sys.exit(1)
+        fields["ucomm"] = len(fields)
+        for line in out:
+            ps_info = line.strip().split(None, len(fields)-2)
+            if "stime" in fields:
+                if ps_info[fields["stime"]] in ["Jan","Feb","Mar","Apr","May","Jun",
+                                           "Jui","Aug","Sep","Oct","Nov","Dec"]:
+                    ps_info = line.strip().split(None, len(fields)-1)
+                    ps_info[fields["stime"]] += ps_info.pop(fields["stime"]+1)
+            ps_info.append(os.path.basename(ps_info[fields["args"]].split()[0]))
+            ps_out.append(' '.join(
+                [('%-'+ str(widths[i]) +'s') % ps_info[fields[opt]]
+                 for i,opt in enumerate(self.ps_fields)] + [ps_info[fields["args"]]]
+            ))
+        return ps_out
 
     def get_psinfo(self, pid_zero):
         """parse unix ps command"""
-
-        ps_cmd = 'ps -e ' + ' '.join(
-                    ['-o ' + o + '=' + 130*'-' for o in self.ps_fields]
-                ) + ' -o args'
-        # print(ps_cmd)
-        ps_out = runcmd(ps_cmd.split(' ')).split('\n')
+        widths = [30, 30, 30, 130] + [50 for i in self.ps_fields[4:]]
+        ps_out = self.run_ps(widths)
         pid_z = ["0", "0"] + self.ps_fields[2:]
-        ps_out[0] = ' '.join(['%-130s' % opt for opt in pid_z] + ['args'])
+        ps_out[0] = ' '.join(
+                [('%-'+ str(widths[i]) +'s') % opt for i,opt in enumerate(pid_z)] + ['args']
+        )
         ps_opts = ['pid', 'ppid', 'user', 'comm'] + self.ps_fields[4:]
-        # print(ps_out[0])
         for line in ps_out:
-            # print(line)
             infos = {}
+            col = 0
             for i,field in enumerate(ps_opts):
-                infos[field] = line[i*131:i*131+130].strip()
-            infos['args'] = line[len(ps_opts)*131:len(line)]
+                infos[field] = line[col:col+widths[i]].strip()
+                col = col + widths[i] + 1
+            infos['args'] = line[col:len(line)]
             infos['comm'] = os.path.basename(infos['comm'])
-            # print(infos)
             pid = infos['pid']
             ppid = infos['ppid']
             if pid == str(os.getpid()):
@@ -182,6 +236,8 @@ class Proctree:
                 self.children[ppid] = []
             self.children[ppid].append(pid)
             self.ps_info[pid] = infos
+        if not self.ps_info.get('1'):
+            self.ps_info['1'] = self.ps_info['0']
         if not pid_zero:
             del self.ps_info['0']
             del self.children['0']
@@ -190,7 +246,7 @@ class Proctree:
         """mini built-in pgrep if pgrep command not available
            [-f] [-x] [-i] [-u <user>] [pattern]"""
         if "PGT_PGREP" not in os.environ or os.environ["PGT_PGREP"]:
-            pgrep = runcmd(['pgrep'] + argv)
+            _, pgrep = runcmd('pgrep ' +' '.join(argv))
             return pgrep.split("\n")
 
         try:
@@ -229,6 +285,7 @@ class Proctree:
 
     def get_parents(self):
         """get parents list of pids"""
+        last_ppid = None
         for pid in self.pids:
             if pid not in self.ps_info:
                 continue
@@ -352,8 +409,42 @@ def wrap_text(opt):
         after = ''
     return after
 
+def pgtree(options, psfields, pgrep_args):
+    """ Display process tree from options """
+    ptree = Proctree(use_uid='-I' in options,
+                     use_ascii='-a' in options,
+                     use_color=colored(options['-C']),
+                     pid_zero='-1' not in options,
+                     opt_fields=psfields,
+                     threads='-T' in options)
+
+    found = None
+    if '-p' in options:
+        found = options['-p'].split(',')
+    elif pgrep_args:
+        found = ptree.pgrep(pgrep_args)
+    return (ptree, found)
+
+def watch_pgtree(options, psfields, pgrep_args, sig):
+    """ follow process hierarchy """
+    while True:
+        try:
+            (ptree, found) = pgtree(options, psfields, pgrep_args)
+            cur_time = time.strftime("%c", time.localtime())
+            sys.stdout.write("\033c")
+            wrap_text(options['-w'])
+            sys.stdout.write("Every 2.0s: " + ' '.join(sys.argv) + "    " + cur_time + "\n\n")
+            ptree.print_tree(pids=found, child_only='-c' in options, sig=sig,
+                     confirmed='-y' in options)
+            if time.sleep(2):
+                break
+        except KeyboardInterrupt:
+            break
+
+
 def main(argv):
     """pgtree command line"""
+    global PS_OPTION
     usage = """
     usage: pgtree.py [-W] [-RIya] [-C <when>] [-O <psfield>] [-c|-k|-K] [-1|-p <pid1>,...|<pgrep args>]
 
@@ -366,8 +457,9 @@ def main(argv):
     -R : force use of internal pgrep
     -C : color preference : y/yes/always or n/no/never (default auto)
     -w : tty wrap text : y/yes or n/no (default y)
-    -W : use watch utility to execute pgtree with default interval
+    -W : watch and follow process tree every 2s
     -a : use ascii characters
+    -T : display threads (ps -T)
     -O <psfield>[,psfield,...] : display multiple <psfield> instead of 'stime' in output
                    <psfield> must be valid with ps -o <psfield> command
 
@@ -387,7 +479,7 @@ def main(argv):
         argv = os.environ["PGTREE"].split(' ') + argv
     try:
         opts, args = getopt.getopt(argv,
-                                   "1IRckKfxvinoyap:u:U:g:G:P:s:t:F:O:C:w:",
+                                   "W1IRckKfxvinoyaTp:u:U:g:G:P:s:t:F:O:C:w:",
                                    ["ns=", "nslist="])
     except getopt.GetoptError:
         print(usage)
@@ -395,7 +487,6 @@ def main(argv):
 
     sig = 0
     pgrep_args = []
-    found = None
     options = {}
     psfields = None
     options['-C'] = 'auto'
@@ -406,30 +497,24 @@ def main(argv):
             sig = 15
         elif opt == "-K":
             sig = 9
-        elif opt == "-p":
-            found = arg.split(',')
         elif opt == "-O":
-            psfields = arg
+            psfields = arg.split(',')
         elif opt == "-R":
             os.environ["PGT_PGREP"] = ""
+        elif opt == "-T":
+            PS_OPTION += " -T"
         elif opt in ("-f", "-x", "-v", "-i", "-n", "-o"):
             pgrep_args.append(opt)
         elif opt in ("-u", "-U", "-g", "-G", "-P", "-s", "-t", "-F", "--ns", "--nslist"):
             pgrep_args += [opt, arg]
     pgrep_args += args
-
     after = wrap_text(options['-w'])
-
-    ptree = Proctree(use_uid='-I' in options,
-                     use_ascii='-a' in options,
-                     use_color=colored(options['-C']),
-                     pid_zero='-1' not in options,
-                     optfields=psfields)
-
-    if pgrep_args:
-        found = ptree.pgrep(pgrep_args)
-    ptree.print_tree(pids=found, child_only='-c' in options, sig=sig,
-                     confirmed='-y' in options)
+    if '-W' in options:
+        watch_pgtree(options, psfields, pgrep_args, sig)
+    else:
+        (ptree, found) = pgtree(options, psfields, pgrep_args)
+        ptree.print_tree(pids=found, child_only='-c' in options, sig=sig,
+                         confirmed='-y' in options)
     sys.stdout.write(after)
 
 
